@@ -106,6 +106,15 @@ const  int64 kint64min =  ( ((( int64) kint32min) << 32) | 0 );
 #define PRIxPTR "lx"
 #endif
 
+// Also allow for printing of a pthread_t.
+#define GPRIuPTHREAD "lu"
+#define GPRIxPTHREAD "lx"
+#if defined(__CYGWIN__) || defined(__CYGWIN32__) || defined(__APPLE__)
+#define PRINTABLE_PTHREAD(pthreadt) reinterpret_cast<uintptr_t>(pthreadt)
+#else
+#define PRINTABLE_PTHREAD(pthreadt) pthreadt
+#endif
+
 // A macro to disallow the evil copy constructor and operator= functions
 // This should be used in the private: declarations for a class
 #define DISALLOW_EVIL_CONSTRUCTORS(TypeName)    \
@@ -231,7 +240,7 @@ struct CompileAssert {
 # define HAVE_ATTRIBUTE_SECTION_START 1
 
 #elif defined(HAVE___ATTRIBUTE__) && defined(__MACH__)
-# define ATTRIBUTE_SECTION(name) __attribute__ ((section ("__DATA, " #name)))
+# define ATTRIBUTE_SECTION(name) __attribute__ ((section ("__TEXT, " #name)))
 
 #include <mach-o/getsect.h>
 #include <mach-o/dyld.h>
@@ -242,18 +251,32 @@ class AssignAttributeStartEnd {
     if (_dyld_present()) {
       for (int i = _dyld_image_count() - 1; i >= 0; --i) {
         const mach_header* hdr = _dyld_get_image_header(i);
-        uint32_t len;
-        *pstart = getsectdatafromheader(hdr, "__DATA", name, &len);
-        if (*pstart) {   // NULL if not defined in this dynamic library
-          *pstart += _dyld_get_image_vmaddr_slide(i);   // correct for reloc
-          *pend = *pstart + len;
-          return;
+#ifdef MH_MAGIC_64
+        if (hdr->magic == MH_MAGIC_64) {
+          uint64_t len;
+          *pstart = getsectdatafromheader_64((mach_header_64*)hdr,
+                                             "__TEXT", name, &len);
+          if (*pstart) {   // NULL if not defined in this dynamic library
+            *pstart += _dyld_get_image_vmaddr_slide(i);   // correct for reloc
+            *pend = *pstart + len;
+            return;
+          }
+        }
+#endif
+        if (hdr->magic == MH_MAGIC) {
+          uint32_t len;
+          *pstart = getsectdatafromheader(hdr, "__TEXT", name, &len);
+          if (*pstart) {   // NULL if not defined in this dynamic library
+            *pstart += _dyld_get_image_vmaddr_slide(i);   // correct for reloc
+            *pend = *pstart + len;
+            return;
+          }
         }
       }
     }
     // If we get here, not defined in a dll at all.  See if defined statically.
     unsigned long len;    // don't ask me why this type isn't uint32_t too...
-    *pstart = getsectdata("__DATA", name, &len);
+    *pstart = getsectdata("__TEXT", name, &len);
     *pend = *pstart + len;
   }
 };
