@@ -128,10 +128,91 @@
   #define ANNOTATE_PUBLISH_MEMORY_RANGE(pointer, size) \
     AnnotatePublishMemoryRange(__FILE__, __LINE__, pointer, size)
 
+  // Report that the bytes in the range [pointer, pointer+size) are not shared
+  // between threads any more and can be safely used by the current thread w/o
+  // synchronization. The race checker will create a happens-before arc from
+  // all previous accesses to this memory to this call.
+  //
+  // This annotation could be applied to complex objects, such as STL
+  // containers, with one condition: the accesses to the object itself
+  // and its internal data should not be separated with any synchronization.
+  //
+  // Example that works:
+  //
+  // map<int, int> the_map;
+  // void Thread1() {
+  //   MutexLock lock(&mu);
+  //   // Ok: accesses to the_map and its internal data is not separated by
+  //   // synchronization.
+  //   the_map[1]++;
+  // }
+  // void Thread2() {
+  //   {
+  //     MutexLock lock(&mu);
+  //     ...
+  //     // because of some reason we know that the_map will not be used by
+  //     // other threads any more
+  //     ANNOTATE_UNPUBLISH_MEMORY_RANGE(&the_map, sizeof(the_map));
+  //   }
+  //   the_map->DoSomething();
+  // }
+  //
+  // Example that does not work (due to the way happens-before arcs are
+  // represented in some race detectors):
+  //
+  // void Thread1() {
+  //   MutexLock lock(&mu);
+  //   int *guts_of_the_map = &(*the_map)[1];
+  //   // we have some synchronization between access to 'c' and its guts.
+  //   // This will make ANNOTATE_UNPUBLISH_MEMORY_RANGE in Thread2  useless.
+  //   some_other_lock_or_other_synchronization_utility.Lock();
+  //   (*guts_of_the_map)++;
+  //    ...
+  // }
+  //
+  // void Thread1() { // same as above...
+  #define ANNOTATE_UNPUBLISH_MEMORY_RANGE(pointer, size) \
+    AnnotateUnpublishMemoryRange(__FILE__, __LINE__, pointer, size)
+
+  // This annotation should be used to annotate thread-safe swapping of
+  // containers. Required only when using hybrid (i.e. not pure happens-before)
+  // detectors.
+  //
+  // This annotation has the same limitation as ANNOTATE_UNPUBLISH_MEMORY_RANGE
+  // (see above).
+  //
+  // Example:
+  // map<int, int> the_map;
+  // void Thread1() {
+  //   MutexLock lock(&mu);
+  //   the_map[1]++;
+  // }
+  // void Thread2() {
+  //   map<int,int> tmp;
+  //   {
+  //     MutexLock lock(&mu);
+  //     the_map.swap(tmp);
+  //     ANNOTATE_SWAP_MEMORY_RANGE(&the_map, sizeof(the_map));
+  //   }
+  //   tmp->DoSomething();
+  // }
+  #define ANNOTATE_SWAP_MEMORY_RANGE(pointer, size)   \
+    do {                                              \
+      ANNOTATE_UNPUBLISH_MEMORY_RANGE(pointer, size); \
+      ANNOTATE_PUBLISH_MEMORY_RANGE(pointer, size);   \
+    } while (0)
+
   // Instruct the tool to create a happens-before arc between mu->Unlock() and
-  // mu->Lock().  This annotation may slow down the race detector; normally it
-  // is used only when it would be difficult to annotate each of the mutex's
-  // critical sections individually using the annotations above.
+  // mu->Lock(). This annotation may slow down the race detector and hide real
+  // races. Normally it is used only when it would be difficult to annotate each
+  // of the mutex's critical sections individually using the annotations above.
+  // This annotation makes sense only for hybrid race detectors. For pure
+  // happens-before detectors this is a no-op. For more details see
+  // http://code.google.com/p/data-race-test/wiki/PureHappensBeforeVsHybrid .
+  #define ANNOTATE_PURE_HAPPENS_BEFORE_MUTEX(mu) \
+    AnnotateMutexIsUsedAsCondVar(__FILE__, __LINE__, mu)
+
+  // Deprecated. Use ANNOTATE_PURE_HAPPENS_BEFORE_MUTEX.
   #define ANNOTATE_MUTEX_IS_USED_AS_CONDVAR(mu) \
     AnnotateMutexIsUsedAsCondVar(__FILE__, __LINE__, mu)
 
@@ -274,7 +355,8 @@
   #define ANNOTATE_HAPPENS_BEFORE(obj) // empty
   #define ANNOTATE_HAPPENS_AFTER(obj) // empty
   #define ANNOTATE_PUBLISH_MEMORY_RANGE(address, size) // empty
-  #define ANNOTATE_PUBLISH_OBJECT(address) // empty
+  #define ANNOTATE_UNPUBLISH_MEMORY_RANGE(address, size)  // empty
+  #define ANNOTATE_SWAP_MEMORY_RANGE(address, size)  // empty
   #define ANNOTATE_PCQ_CREATE(pcq) // empty
   #define ANNOTATE_PCQ_DESTROY(pcq) // empty
   #define ANNOTATE_PCQ_PUT(pcq) // empty
@@ -282,6 +364,7 @@
   #define ANNOTATE_NEW_MEMORY(address, size) // empty
   #define ANNOTATE_EXPECT_RACE(address, description) // empty
   #define ANNOTATE_BENIGN_RACE(address, description) // empty
+  #define ANNOTATE_PURE_HAPPENS_BEFORE_MUTEX(mu) // empty
   #define ANNOTATE_MUTEX_IS_USED_AS_CONDVAR(mu) // empty
   #define ANNOTATE_TRACE_MEMORY(arg) // empty
   #define ANNOTATE_THREAD_NAME(name) // empty
@@ -314,6 +397,9 @@ extern "C" void AnnotateCondVarSignalAll(const char *file, int line,
 extern "C" void AnnotatePublishMemoryRange(const char *file, int line,
                                            const volatile void *address,
                                            long size);
+extern "C" void AnnotateUnpublishMemoryRange(const char *file, int line,
+                                           const volatile void *address,
+                                           long size);
 extern "C" void AnnotatePCQCreate(const char *file, int line,
                                   const volatile void *pcq);
 extern "C" void AnnotatePCQDestroy(const char *file, int line,
