@@ -1,10 +1,10 @@
 // Copyright (c) 2005, Google Inc.
 // All rights reserved.
-// 
+//
 // Redistribution and use in source and binary forms, with or without
 // modification, are permitted provided that the following conditions are
 // met:
-// 
+//
 //     * Redistributions of source code must retain the above copyright
 // notice, this list of conditions and the following disclaimer.
 //     * Redistributions in binary form must reproduce the above
@@ -14,7 +14,7 @@
 //     * Neither the name of Google Inc. nor the names of its
 // contributors may be used to endorse or promote products derived from
 // this software without specific prior written permission.
-// 
+//
 // THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 // "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 // LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -33,7 +33,7 @@
 // Author: Maxim Lifantsev
 //
 
-#include <config.h>
+#include "config.h"
 
 #include <fcntl.h>    // for O_RDONLY (we use syscall to do actual reads)
 #include <string.h>
@@ -47,9 +47,9 @@
 #ifdef HAVE_PTHREAD
 #include <pthread.h>
 #endif
-#include <sys/poll.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <time.h>
 #include <assert.h>
 
 #ifdef HAVE_LINUX_PTRACE_H
@@ -1489,7 +1489,7 @@ void HeapLeakChecker::Create(const char *name, bool make_start_snapshot) {
     } else {
       RAW_LOG(WARNING, "Heap checker is not active, "
                        "hence checker \"%s\" will do nothing!", name);
-    RAW_LOG(WARNING, "To activate set the HEAPCHECK environment variable.\n");	
+    RAW_LOG(WARNING, "To activate set the HEAPCHECK environment variable.\n");
     }
   }
   if (name_ == NULL) {
@@ -1576,7 +1576,11 @@ static void SuggestPprofCommand(const char* pprof_file_arg) {
           "If you are still puzzled about why the leaks are "
           "there, try rerunning this program with "
           "HEAP_CHECK_TEST_POINTER_ALIGNMENT=1 and/or with "
-          "HEAP_CHECK_MAX_POINTER_OFFSET=-1\n",
+          "HEAP_CHECK_MAX_POINTER_OFFSET=-1\n"
+          "If the leak report occurs in a small fraction of runs, "
+          "try running with TCMALLOC_MAX_FREE_QUEUE_SIZE of few hundred MB "
+          "or with TCMALLOC_RECLAIM_MEMORY=false, "  // only works for debugalloc
+          "it might help find leaks more repeatably\n",
           fetch_cmd.c_str(),
           "pprof",           // works as long as pprof is on your path
           invocation_path().c_str(),
@@ -1715,11 +1719,16 @@ bool HeapLeakChecker::DoNoLeaks(ShouldSymbolize should_symbolize) {
       // this makes it unsafe to ever leak-report again!  Luckily, we
       // typically only want to report once in a program's run, at the
       // very end.
-      CancelInitialMallocHooks();
       if (MallocHook::GetNewHook() == NewHook)
         MallocHook::SetNewHook(NULL);
       if (MallocHook::GetDeleteHook() == DeleteHook)
         MallocHook::SetDeleteHook(NULL);
+      MemoryRegionMap::Shutdown();
+      // Make sure all the hooks really got unset:
+      RAW_CHECK(MallocHook::GetNewHook() == NULL, "");
+      RAW_CHECK(MallocHook::GetDeleteHook() == NULL, "");
+      RAW_CHECK(MallocHook::GetMmapHook() == NULL, "");
+      RAW_CHECK(MallocHook::GetSbrkHook() == NULL, "");
       have_disabled_hooks_for_symbolize = true;
       leaks->ReportLeaks(name_, pprof_file, true);  // true = should_symbolize
     } else {
@@ -1823,7 +1832,9 @@ static bool internal_init_start_has_run = false;
 //
 /*static*/ void HeapLeakChecker::InternalInitStart() {
   { SpinLockHolder l(&heap_checker_lock);
-    RAW_CHECK(!internal_init_start_has_run, "Only one call is expected");
+    RAW_CHECK(!internal_init_start_has_run,
+              "Heap-check constructor called twice.  Perhaps you both linked"
+              " in the heap checker, and also used LD_PRELOAD to load it?");
     internal_init_start_has_run = true;
 
     if (FLAGS_heap_check.empty()) {
@@ -1977,6 +1988,21 @@ static bool internal_init_start_has_run = false;
     RAW_CHECK(heap_checker_on  &&  constructor_heap_profiling,
               "Leak checking is expected to be fully turned on now");
   }
+
+  // For binaries built in debug mode, this will set release queue of
+  // debugallocation.cc to 100M to make it less likely for real leaks to
+  // be hidden due to reuse of heap memory object addresses.
+  // Running a test with --malloc_reclaim_memory=0 would help find leaks even
+  // better, but the test might run out of memory as a result.
+  // The scenario is that a heap object at address X is allocated and freed,
+  // but some other data-structure still retains a pointer to X.
+  // Then the same heap memory is used for another object, which is leaked,
+  // but the leak is not noticed due to the pointer to the original object at X.
+  // TODO(csilvers): support this in some manner.
+#if 0
+  SetCommandLineOptionWithMode("max_free_queue_size", "104857600",  // 100M
+                               SET_FLAG_IF_DEFAULT);
+#endif
 }
 
 // We want this to run early as well, but not so early as
@@ -2272,7 +2298,8 @@ void HeapLeakChecker_AfterDestructors() {
   }
   if (FLAGS_heap_check_after_destructors) {
     if (HeapLeakChecker::DoMainHeapCheck()) {
-      poll(0, 0, 500);
+      const struct timespec sleep_time = { 0, 500000000 };  // 500 ms
+      nanosleep(&sleep_time, NULL);
         // Need this hack to wait for other pthreads to exit.
         // Otherwise tcmalloc find errors
         // on a free() call from pthreads.
