@@ -37,12 +37,6 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>   // for write()
 #endif
-#ifdef HAVE_SYS_SOCKET_H
-#include <sys/socket.h>   // for socketpair() -- needed by Symbolize
-#endif
-#ifdef HAVE_SYS_WAIT_H
-#include <sys/wait.h>   // for wait() -- needed by Symbolize
-#endif
 #include <fcntl.h>    // for open()
 #ifdef HAVE_GLOB_H
 #include <glob.h>
@@ -53,6 +47,9 @@
 #ifdef HAVE_INTTYPES_H
 #include <inttypes.h> // for PRIxPTR
 #endif
+#ifdef HAVE_POLL_H
+#include <poll.h>
+#endif
 #include <errno.h>
 #include <stdarg.h>
 #include <string>
@@ -63,6 +60,7 @@
 
 #include "base/logging.h"
 #include "raw_printer.h"
+#include "symbolize.h"
 #include <google/stacktrace.h>
 #include <google/malloc_hook.h>
 #include "base/commandlineflags.h"
@@ -84,14 +82,9 @@ DEFINE_bool(cleanup_old_heap_profiles,
             EnvToBool("HEAP_PROFILE_CLEANUP", true),
             "At initialization time, delete old heap profiles.");
 
-DEFINE_string(heap_profile_table_pprof,
-              EnvToString("PPROF_PATH", "pprof"),
-              "Path to pprof to call for reporting function names.");
-
-// heap_profile_table_pprof may be referenced after destructors are
-// called (since that's when leak-checking is done), so we make
-// a more-permanent copy that won't ever get destroyed.
-static string* g_pprof_path = new string(FLAGS_heap_profile_table_pprof);
+DEFINE_int32(heap_check_max_leaks,
+             EnvToInt("HEAP_CHECK_MAX_LEAKS", 20),
+             "The maximum number of leak reports to print.");
 
 //----------------------------------------------------------------------
 
@@ -521,86 +514,6 @@ void HeapProfileTable::Snapshot::ReportCallback(const void* ptr,
   e->bytes += v->bytes;
 }
 
-// It would be much more efficient to call Symbolize on a bunch of pc's
-// at once, rather than calling one at a time, but this way it's simpler
-// to code, and efficiency probably isn't a top concern when reporting
-// found leaks at program-exit time.
-// Note that the forking/etc is not thread-safe or re-entrant.  That's
-// ok for the purpose we need -- reporting leaks detected by heap-checker
-// -- but be careful if you decide to use this routine for other purposes.
-static bool Symbolize(void *pc, char *out, int out_size) {
-#if !defined(HAVE_UNISTD_H)  || !defined(HAVE_SYS_SOCKET_H) || !defined(HAVE_SYS_WAIT_H)
-  return false;
-#elif !defined(HAVE_PROGRAM_INVOCATION_NAME)
-  return false;   // TODO(csilvers): get argv[0] somehow
-#else
-  // All this work is to do two-way communication.  ugh.
-  extern char* program_invocation_name;  // gcc provides this
-  int child_in[2];   // file descriptors
-  int child_out[2];  // for now, we don't worry about child_err
-  if (socketpair(AF_UNIX, SOCK_STREAM, 0, child_in) == -1) {
-    return false;
-  }
-  if (socketpair(AF_UNIX, SOCK_STREAM, 0, child_out) == -1) {
-    close(child_in[0]);
-    close(child_in[1]);
-    return false;
-  }
-  switch (fork()) {
-    case -1: {  // error
-      close(child_in[0]);
-      close(child_in[1]);
-      close(child_out[0]);
-      close(child_out[1]);
-      return false;
-    }
-    case 0: {  // child
-      close(child_in[1]);   // child uses the 0's, parent uses the 1's
-      close(child_out[1]);  // child uses the 0's, parent uses the 1's
-      close(0);
-      close(1);
-      if (dup2(child_in[0], 0) == -1) _exit(1);
-      if (dup2(child_out[0], 1) == -1) _exit(2);
-      execlp(g_pprof_path->c_str(), g_pprof_path->c_str(),
-             "--symbols", program_invocation_name, NULL);
-      _exit(3);  // if execvp fails, it's bad news for us
-    }
-    default: {  // parent
-      close(child_in[0]);   // child uses the 0's, parent uses the 1's
-      close(child_out[0]);  // child uses the 0's, parent uses the 1's
-      DumpProcSelfMaps(child_in[1]);  // what pprof expects on stdin
-      char pcstr[64];                 // enough for a single address
-      snprintf(pcstr, sizeof(pcstr),  // pprof expects format to be 0xXXXXXX...
-               "0x%" PRIxPTR "\n", reinterpret_cast<uintptr_t>(pc));
-      write(child_in[1], pcstr, strlen(pcstr));
-      close(child_in[1]);             // that's all we need to write
-      int total_bytes_read = 0;
-      while (1) {
-        int bytes_read = read(child_out[1], out + total_bytes_read,
-                              out_size - total_bytes_read);
-        if (bytes_read < 0) {
-          close(child_out[1]);
-          return false;
-        } else if (bytes_read == 0) {
-          close(child_out[1]);
-          wait(NULL);
-          break;
-        } else {
-          total_bytes_read += bytes_read;
-        }
-      }
-      // We have successfully read the output of pprof into out.  Make sure
-      // we got the full symbol (we can tell because it ends with a \n).
-      if (total_bytes_read == 0 || out[total_bytes_read - 1] != '\n')
-        return false;
-      out[total_bytes_read - 1] = '\0';  // remove the trailing newline
-      return true;
-    }
-  }
-  return false;  // shouldn't be reachable
-#endif
-}
-
 void HeapProfileTable::Snapshot::ReportLeaks(const char* checker_name,
                                              const char* filename,
                                              bool should_symbolize) {
@@ -630,13 +543,23 @@ void HeapProfileTable::Snapshot::ReportLeaks(const char* checker_name,
 
   // Report a bounded number of leaks to keep the leak report from
   // growing too long.
-  const int to_report = (n > 20) ? 20 : n;
+  const int to_report =
+      (FLAGS_heap_check_max_leaks > 0 &&
+       n > FLAGS_heap_check_max_leaks) ? FLAGS_heap_check_max_leaks : n;
   RAW_LOG(ERROR, "The %d largest leaks:", to_report);
 
   // Print
-  char sym_buffer[1024];
+  SymbolTable symbolization_table;
+  for (int i = 0; i < to_report; i++) {
+    const Entry& e = entries[i];
+    for (int j = 0; j < e.bucket->depth; j++) {
+      symbolization_table.Add(e.bucket->stack[j]);
+    }
+  }
   static const int kBufSize = 2<<10;
   char buffer[kBufSize];
+  if (should_symbolize)
+    symbolization_table.Symbolize();
   for (int i = 0; i < to_report; i++) {
     const Entry& e = entries[i];
     base::RawPrinter printer(buffer, kBufSize);
@@ -644,17 +567,12 @@ void HeapProfileTable::Snapshot::ReportLeaks(const char* checker_name,
                    e.bytes, e.count);
     for (int j = 0; j < e.bucket->depth; j++) {
       const void* pc = e.bucket->stack[j];
-      const char* sym;
-      if (should_symbolize &&
-          Symbolize(const_cast<void*>(pc), sym_buffer, sizeof(sym_buffer))) {
-        sym = sym_buffer;
-      } else {
-        sym = "";
-      }
-      printer.Printf("\t@ %p %s\n", pc, sym);
+      printer.Printf("\t@ %"PRIxPTR" %s\n",
+          reinterpret_cast<uintptr_t>(pc), symbolization_table.GetSymbol(pc));
     }
     RAW_LOG(ERROR, "%s", buffer);
   }
+
   if (to_report < n) {
     RAW_LOG(ERROR, "Skipping leaks numbered %d..%d",
             to_report, n-1);
