@@ -45,6 +45,9 @@
 #include <string>
 #include "base/dynamic_annotations.h"
 #include "base/sysinfo.h"    // for FillProcSelfMaps
+#ifndef NO_HEAP_CHECK
+#include "google/heap-checker.h"
+#endif
 #include "google/malloc_extension.h"
 #include "maybe_threads.h"
 
@@ -120,7 +123,7 @@ bool MallocExtension::MallocMemoryStats(int* blocks, size_t* total,
                                        int histogram[kMallocHistogramSize]) {
   *blocks = 0;
   *total = 0;
-  memset(histogram, 0, sizeof(histogram));
+  memset(histogram, 0, sizeof(*histogram) * kMallocHistogramSize);
   return true;
 }
 
@@ -136,10 +139,18 @@ void MallocExtension::MarkThreadIdle() {
   // Default implementation does nothing
 }
 
-void MallocExtension::ReleaseFreeMemory() {
+void MallocExtension::MarkThreadBusy() {
   // Default implementation does nothing
 }
 
+void MallocExtension::ReleaseToSystem(size_t num_bytes) {
+  // Default implementation does nothing
+}
+
+void MallocExtension::ReleaseFreeMemory() {
+  ReleaseToSystem(static_cast<size_t>(-1));   // SIZE_T_MAX
+}
+
 void MallocExtension::SetMemoryReleaseRate(double rate) {
   // Default implementation does nothing
 }
@@ -156,17 +167,16 @@ size_t MallocExtension::GetAllocatedSize(void* p) {
   return 0;
 }
 
-// The current malloc extension object.  We also keep a pointer to
-// the default implementation so that the heap-leak checker does not
-// complain about a memory leak.
+// The current malloc extension object.
 
 static pthread_once_t module_init = PTHREAD_ONCE_INIT;
-static MallocExtension* default_instance = NULL;
 static MallocExtension* current_instance = NULL;
 
 static void InitModule() {
-  default_instance = new MallocExtension;
-  current_instance = default_instance;
+  current_instance = new MallocExtension;
+#ifndef NO_HEAP_CHECK
+  HeapLeakChecker::IgnoreObject(current_instance);
+#endif
 }
 
 MallocExtension* MallocExtension::instance() {
@@ -294,6 +304,10 @@ void MallocExtension::GetHeapGrowthStacks(MallocExtensionWriter* writer) {
   DumpAddressMap(writer);
 }
 
+void MallocExtension::Ranges(void* arg, RangeFunction func) {
+  // No callbacks by default
+}
+
 // These are C shims that work on the current instance.
 
 #define C_SHIM(fn, retval, paramlist, arglist)          \
@@ -301,22 +315,24 @@ void MallocExtension::GetHeapGrowthStacks(MallocExtensionWriter* writer) {
     return MallocExtension::instance()->fn arglist;     \
   }
 
-C_SHIM(VerifyAllMemory, bool, (), ());
-C_SHIM(VerifyNewMemory, bool, (void* p), (p));
-C_SHIM(VerifyArrayNewMemory, bool, (void* p), (p));
-C_SHIM(VerifyMallocMemory, bool, (void* p), (p));
-C_SHIM(MallocMemoryStats, bool,
+C_SHIM(VerifyAllMemory, int, (void), ());
+C_SHIM(VerifyNewMemory, int, (void* p), (p));
+C_SHIM(VerifyArrayNewMemory, int, (void* p), (p));
+C_SHIM(VerifyMallocMemory, int, (void* p), (p));
+C_SHIM(MallocMemoryStats, int,
        (int* blocks, size_t* total, int histogram[kMallocHistogramSize]),
        (blocks, total, histogram));
 
 C_SHIM(GetStats, void,
        (char* buffer, int buffer_length), (buffer, buffer_length));
-C_SHIM(GetNumericProperty, bool,
+C_SHIM(GetNumericProperty, int,
        (const char* property, size_t* value), (property, value));
-C_SHIM(SetNumericProperty, bool,
+C_SHIM(SetNumericProperty, int,
        (const char* property, size_t value), (property, value));
 
-C_SHIM(MarkThreadIdle, void, (), ());
-C_SHIM(ReleaseFreeMemory, void, (), ());
+C_SHIM(MarkThreadIdle, void, (void), ());
+C_SHIM(MarkThreadBusy, void, (void), ());
+C_SHIM(ReleaseFreeMemory, void, (void), ());
+C_SHIM(ReleaseToSystem, void, (size_t num_bytes), (num_bytes));
 C_SHIM(GetEstimatedAllocatedSize, size_t, (size_t size), (size));
 C_SHIM(GetAllocatedSize, size_t, (void* p), (p));
