@@ -42,7 +42,6 @@ extern "C" {
 }
 #include "google/stacktrace.h"
 #include "base/logging.h"
-#include "base/spinlock.h"
 
 // Sometimes, we can try to get a stack trace from within a stack
 // trace, because libunwind can call mmap (maybe indirectly via an
@@ -51,7 +50,7 @@ extern "C" {
 // recursive request, we'd end up with infinite recursion or deadlock.
 // Luckily, it's safe to ignore those subsequent traces.  In such
 // cases, we return 0 to indicate the situation.
-static SpinLock libunwind_lock(SpinLock::LINKER_INITIALIZED);
+static __thread int recursive;
 
 // If you change this function, also change GetStackFrames below.
 int GetStackTrace(void** result, int max_depth, int skip_count) {
@@ -60,9 +59,10 @@ int GetStackTrace(void** result, int max_depth, int skip_count) {
   unw_cursor_t cursor;
   unw_context_t uc;
 
-  if (!libunwind_lock.TryLock()) {
+  if (recursive) {
     return 0;
   }
+  ++recursive;
 
   unw_getcontext(&uc);
   int ret = unw_init_local(&cursor, &uc);
@@ -70,20 +70,19 @@ int GetStackTrace(void** result, int max_depth, int skip_count) {
   skip_count++;         // Do not include the "GetStackTrace" frame
 
   while (n < max_depth) {
-    int ret = unw_get_reg(&cursor, UNW_REG_IP, (unw_word_t *) &ip);
-    if (ret < 0)
+    if (unw_get_reg(&cursor, UNW_REG_IP, (unw_word_t *) &ip) < 0) {
       break;
+    }
     if (skip_count > 0) {
       skip_count--;
     } else {
       result[n++] = ip;
     }
-    ret = unw_step(&cursor);
-    if (ret <= 0)
+    if (unw_step(&cursor) <= 0) {
       break;
+    }
   }
-
-  libunwind_lock.Unlock();
+  --recursive;
   return n;
 }
 
@@ -121,9 +120,10 @@ int GetStackFrames(void** pcs, int* sizes, int max_depth, int skip_count) {
   unw_context_t uc;
   unw_word_t sp = 0, next_sp = 0;
 
-  if (!libunwind_lock.TryLock()) {
+  if (recursive) {
     return 0;
   }
+  ++recursive;
 
   unw_getcontext(&uc);
   RAW_CHECK(unw_init_local(&cursor, &uc) >= 0, "unw_init_local failed");
@@ -151,6 +151,6 @@ int GetStackFrames(void** pcs, int* sizes, int max_depth, int skip_count) {
     pcs[n++] = ip;
   }
  out:
-  libunwind_lock.Unlock();
+  --recursive;
   return n;
 }
