@@ -1,10 +1,10 @@
 // Copyright (c) 2005, Google Inc.
 // All rights reserved.
-// 
+//
 // Redistribution and use in source and binary forms, with or without
 // modification, are permitted provided that the following conditions are
 // met:
-// 
+//
 //     * Redistributions of source code must retain the above copyright
 // notice, this list of conditions and the following disclaimer.
 //     * Redistributions in binary form must reproduce the above
@@ -14,7 +14,7 @@
 //     * Neither the name of Google Inc. nor the names of its
 // contributors may be used to endorse or promote products derived from
 // this software without specific prior written permission.
-// 
+//
 // THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 // "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 // LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
@@ -58,7 +58,9 @@
 // (see the comment in our .h file).
 
 #include "config_for_unittests.h"
-#include <sys/poll.h>
+#ifdef HAVE_POLL_H
+#include <poll.h>
+#endif
 #if defined HAVE_STDINT_H
 #include <stdint.h>             // to get uint16_t (ISO naming madness)
 #elif defined HAVE_INTTYPES_H
@@ -384,6 +386,25 @@ static void DeAllocHidden(void** ptr) {
   Use(ptr);
 }
 
+void PreventHeapReclaiming(size_t size) {
+#ifdef NDEBUG
+  if (true) {
+    static void** no_reclaim_list = NULL;
+    CHECK(size >= sizeof(void*));
+    // We can't use malloc_reclaim_memory flag in opt mode as debugallocation.cc
+    // is not used. Instead we allocate a bunch of heap objects that are
+    // of the same size as what we are going to leak to ensure that the object
+    // we are about to leak is not at the same address as some old allocated
+    // and freed object that might still have pointers leading to it.
+    for (int i = 0; i < 100; ++i) {
+      void** p = reinterpret_cast<void**>(new(initialized) char[size]);
+      p[0] = no_reclaim_list;
+      no_reclaim_list = p;
+    }
+  }
+#endif
+}
+
 static bool RunSilent(HeapLeakChecker* check,
                       bool (HeapLeakChecker::* func)()) {
   // By default, don't print the 'we detected a leak' message in the
@@ -436,6 +457,7 @@ static void TestHeapLeakCheckerDeathSimple() {
 }
 
 static void MakeDeathLoop(void** arr1, void** arr2) {
+  PreventHeapReclaiming(2 * sizeof(void*));
   void** a1 = new(initialized) void*[2];
   void** a2 = new(initialized) void*[2];
   a1[1] = reinterpret_cast<void*>(a2);
@@ -539,6 +561,20 @@ static void TestHeapLeakCheckerDeathCountMore() {
   DeAllocHidden(&bar2);
 }
 
+static void TestHiddenPointer() {
+  int i;
+  void* foo = &i;
+  HiddenPointer<void> p(foo);
+  CHECK_EQ(foo, p.get());
+
+  // Confirm pointer doesn't appear to contain a byte sequence
+  // that == the pointer.  We don't really need to test that
+  // the xor trick itself works, as without it nothing in this
+  // test suite would work.  See the Hide/Unhide/*Hidden* set
+  // of helper methods.
+  CHECK_NE(foo, *reinterpret_cast<void**>(&p));
+}
+
 // simple tests that deallocate what they allocated
 static void TestHeapLeakChecker() {
   { HeapLeakChecker check("trivial");
@@ -1294,6 +1330,13 @@ extern void VerifyHeapProfileTableStackGet() {
 
 // ========================================================================= //
 
+static void MakeALeak(void** arr) {
+  PreventHeapReclaiming(10 * sizeof(int));
+  void* a = new(initialized) int[10];
+  Hide(&a);
+  *arr = a;
+}
+
 // Helper to do 'return 0;' inside main(): insted we do 'return Pass();'
 static int Pass() {
   fprintf(stdout, "PASS\n");
@@ -1304,7 +1347,6 @@ static int Pass() {
 int main(int argc, char** argv) {
   run_hidden_ptr = DoRunHidden;
   wipe_stack_ptr = DoWipeStack;
-  
   if (!HeapLeakChecker::IsActive()) {
     CHECK_EQ(FLAGS_heap_check, "");
     LOG(WARNING, "HeapLeakChecker got turned off; we won't test much...");
@@ -1337,7 +1379,8 @@ int main(int argc, char** argv) {
   CHECK(HeapLeakChecker::NoGlobalLeaks());  // so far, so good
 
   if (FLAGS_test_leak) {
-    void* arr = AllocHidden(10 * sizeof(int));
+    void* arr;
+    RunHidden(NewCallback(MakeALeak, &arr));
     Use(&arr);
     LogHidden("Leaking", arr);
     if (FLAGS_test_cancel_global_check)  HeapLeakChecker::CancelGlobalCheck();
@@ -1376,6 +1419,8 @@ int main(int argc, char** argv) {
 
   HeapLeakChecker heap_check("all");
 
+  TestHiddenPointer();
+
   TestHeapLeakChecker();
   Pause();
   TestLeakButTotalsMatch();
@@ -1443,6 +1488,6 @@ int main(int argc, char** argv) {
   }
 
   CHECK(HeapLeakChecker::NoGlobalLeaks());  // so far, so good
-  
+
   return Pass();
 }
