@@ -182,8 +182,6 @@ pthread_key_t PthreadKeyCreate(void (*destr_fn)(void*)) {
 // -----------------------------------------------------------------------
 // These functions replace system-alloc.cc
 
-static SpinLock alloc_lock(SpinLock::LINKER_INITIALIZED);
-
 // This is mostly like MmapSysAllocator::Alloc, except it does these weird
 // munmap's in the middle of the page, which is forbidden in windows.
 extern void* TCMalloc_SystemAlloc(size_t size, size_t *actual_size,
@@ -193,7 +191,6 @@ extern void* TCMalloc_SystemAlloc(size_t size, size_t *actual_size,
     *actual_size = size;
   }
 
-  SpinLockHolder sh(&alloc_lock);
   // Align on the pagesize boundary
   const int pagesize = getpagesize();
   if (alignment < pagesize) alignment = pagesize;
